/* ZX Spectrum mathematical functions

   The functions use the same algorithms and Chebyshev polynomials of degree 6, 8, or 12 as implemented
   in the ZX Spectrum ROM, with similar accuracy.

   All functions are based on the book `The Complete Sectrum ROM Disassembly`, written by Dr. Ian Logan &
   Dr. Frank OHara, pp. 217, available - among other sites - at ftp://ftp.worldofspectrum.org/pub/sinclair/books.

   In general, the procedures are mostly slower and also less precise than their Agena pendants. By default, the fully
   expanded and simplified polynomials are hard-wired into the C code. By passing the optional last argument `true`,
   however, the polynomials are processed iteratively in real-time, using the `zx.genseries` function which imitates
   the Z80 assembler subprocedure `series generator`. You may query the respective Chebyshev coefficient vectors by
   calling `zx.getcoeffs`, and globally change them with `zx.setcoeffs`.

   The C source file src/zx.c contains exact information on the precision of the functions.

   Initiated November 11, 2015; Alexander Walz.

   For licence and copyright, see agena.h file.

   See also article `Polynomial Approximations to Elementary Functions` by C. W. Clenshaw, National Physical Laboratory,
   Teddington, Middlesex, England. */

#include <stdlib.h>
#include <math.h>
#include <string.h>

#define zx_c
#define LUA_LIB

#include "agena.h"

#include "agnxlib.h"
#include "agenalib.h"
#include "agnhlps.h"
#include "agnconf.h"

#include "agncmpt.h"

#define AGENA_LIBVERSION	"zx 0.0.1 for Agena as of November 15, 2015\n"


#if !(defined(LUA_DOS) || defined(__OS2__) || defined(LUA_ANSI))
#define AGENA_ZXLIBNAME "zx"
LUALIB_API int (luaopen_zx) (lua_State *L);
#endif


/* Chebyshev coefficients for zx.ATN (and indirectly zx.ACS, zx.ASN) */
size_t ATNA_N = 12;

lua_Number ATNA[256] = {
  -0.0000000002, 0.0000000010, -0.0000000066, 0.0000000432, -0.0000002850, 0.0000019105,
  -0.0000131076, 0.0000928715, -0.0006905975, 0.0055679210, -0.0529464623, 0.8813735870,
  /* defaults */
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};


/* Chebyshev coefficients for zx.SIN (and zx.COS) */
size_t SINA_N = 6;

lua_Number SINA[256] = {
  -0.000000003, 0.000000592, -0.000068294, 0.004559008, -0.142630785, 1.276278962,
  /* defaults */
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};


/* Chebyshev coefficients for zx.LN (also used by zx.SQR and zx.POW) */
size_t LNA_N = 12;

lua_Number LNA[256] = {
  -0.0000000003, 0.0000000020, -0.0000000127, -0.0000000823, -0.0000005389, 0.0000035828,
  -0.0000243013, 0.0001693953, -0.0012282837, 0.0094766116,  -0.0818414567, 0.9302292213,
  /* defaults */
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};


/* Chebyshev coefficients for zx.EXP (also used by zx.SQR and zx.POW) */
size_t EXPA_N = 8;

lua_Number EXPA[256] = {
  0.000000001, 0.000000053, 0.000001851, 0.000053453, 0.001235714, 0.021446556, 0.248762434, 1.456999875,
  /* defaults */
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};


/* Receives a number z in the range [-1, 1] and a sequence of coefficients and returns the value of the corresponding
   Chebyshev polynomial. If z is out of range, _no_ error is returned. */

lua_Number genseries (lua_Number z, lua_Number *a, size_t breg) {
  lua_Number m0, m1, m2, t, u;
  int i;
  m0 = 2*z;
  m1 = m2 = t = 0;
  for (i=breg; i > 0; i--) {
    m1 = m2;
    u = t*m0 - m2 + a[breg - i];
    m2 = t;
    t = u;
  }
  return t - m1;
}

static int zx_genseries (lua_State *L) {
  lua_Number z, *a;
  int breg, i;
  z = agn_checknumber(L, 1);
  luaL_argcheck(L, lua_isseq(L, 2), 2, "sequence expected");
  breg = agn_seqsize(L, 2);
  if (breg < 1)
    luaL_error(L, "Error in " LUA_QS ": at least one coefficient expected.", "zx.genseries");
  a = malloc(breg * sizeof(lua_Number));
  if (a == NULL)
    luaL_error(L, "Error in " LUA_QS ": memory allocation error.", "zx.genseries");
  for (i=0; i < breg; i++) a[i] = agn_seqgetinumber(L, 2, i);
  lua_pushnumber(L, genseries(z, a, breg));
  xfree(a);
  return 1;
}


/* ZX Spectrum sine function

   Deviations with respect to C's sin:
   median:    1.5157375354846e-010
   mean:      2.4453803557139e-010
   maxmimum:  9.1993013207059e-010

   Reference: for arg from -100 to 100 by 0.01.

   12 % faster than C's sin function. */

#define PSIN(z) (1.267162131 - 0.284851843*z + 0.18226552e-1*z*z - 0.546208e-3*z*z*z + 0.9480e-5*z*z*z*z - 0.112e-6*z*z*z*z*z)

static int zx_sin (lua_State *L) {
  lua_Number x, w, z;
  int gen;
  x = agn_checknumber(L, 1)*0.15915494309189533577;  /* 0.1591549... = 0.5/Pi */
  gen = agnL_optboolean(L, 2, 0);
  x -= floor(x + 0.5);
  w = 4 * x;
  if (w > 1) w = 2 - w;
  else if (w < -1) w = -w - 2;
  z = 2 * w * w - 1;
  lua_pushnumber(L, w*( (gen) ? genseries(z, SINA, SINA_N) : PSIN(z)));
  return 1;
}

/* ZX Spectrum cosine function

   Deviations with respect to C's cos:
   median:    1.5130774411176e-010
   mean:      2.4266713755513e-010
   maxmimum:  9.1993102024901e-010

   Reference: for arg from -100 to 100 by 0.01

   8 % faster than C's cos function. */

static int zx_cos (lua_State *L) {
  lua_Number x, w, z;
  int gen;
  x = agn_checknumber(L, 1)*0.15915494309189533577;  /* 0.15915494309189533577 = 0.5/Pi */
  gen = agnL_optboolean(L, 2, 0);
  x -= floor(x + 0.5);
  w = 4 * x;
  /* this is how the ZX Spectrum 48 ROM assembler routine at offset 20 works: preparation to call sine */
  w = (w > 0) ? 1 - w : -w + 1;
  /* argument transformed, now calculate `sine` */
  if (w > 1) w = 2 - w;
  else if (w < -1) w = -w - 2;
  z = 2 * w * w - 1;
  lua_pushnumber(L, w*( (gen) ? genseries(z, SINA, SINA_N) : PSIN(z)));
  return 1;
}


/* ZX Spectrum tangent function

   Deviations with respect to C's tan:
   median:    2.5561464056523e-010
   mean:      3.2139040809957e-009
   maxmimum:  5.4963111324469e-006

   Reference: for arg from -100 to 100 by 0.01

   7 % faster than C's tan function. */

static int zx_tan (lua_State *L) {
  lua_Number x, w, z, a[2];
  int i, gen;
  x = agn_checknumber(L, 1)*0.15915494309189533577;  /* 0.15915494309189533577 = 0.5/Pi */
  gen = agnL_optboolean(L, 2, 0);
  x -= floor(x + 0.5);
  /* this is how the ZX Spectrum 48 ROM assembler routine at offset 20 works: preparation to call sine */
  for (i=0; i < 2; i++) {
    w = 4 * x;
    if (i == 1) w = (w > 0) ? 1 - w : -w + 1;  /* cosine to be computed ? */
    /* argument transformed, now calculate sine (i == 0) or cosine (i == 1) */
    if (w > 1) w = 2 - w;
    else if (w < -1) w = -w - 2;
    z = 2 * w * w - 1;
    a[i] = w * ((gen) ? genseries(z, SINA, SINA_N) : PSIN(z));
  }
  lua_pushnumber(L, a[0]/a[1]);
  return 1;
}


/* ZX Spectrum ln function (natural logarithm)

   Deviations with respect to C's log:
   median:    3.3824310019526e-008
   mean:      5.0304746414217e-008
   maxmimum:  1.9750573576616e-007

   Reference: for arg from 0.0001 to 100 by 0.0001

   3 % slower than C's log function. */

lua_Number zxln (lua_Number d, int gen) {
  lua_Number s, z, t1, t2, t4, t6;
  int n;
  if (d <= 0) return AGN_NAN;
  d = frexp(d, &n);  /* (N holds e', the exponent */
  if (d > 0.8) {  /* (D holds X'). */
    s = d - 1;
    z = 2.5*d - 3;
  } else {
    n--;
    s = 2*d - 1;
    z = 5*d - 3;
  }
  if (gen)
    z = genseries(z, LNA, LNA_N);
  else {
    t1 = z*z;
    t2 = t1*t1;
    t4 = t2*t2;
    t6 = t1*z;
    z = -0.6671616E-3*t2*z - 0.1565489016*z - 0.48128E-5*t4*z - 0.6144E-6*t4*t6
        - 0.560384E-4*t2*t6 - 0.89116584E-2*t6 + 0.2048E-5*t4*t1 - 0.261888E-4*t4
        + 0.2759168E-3*t2*t1 + 0.233844E-2*t2 + 0.353305696E-1*t1 + 0.9116074545;
  }
  return s*z + n*LN2;

}

static int zx_ln (lua_State *L) {
  lua_pushnumber(L, zxln(agn_checknumber(L, 1), agnL_optboolean(L, 2, 0)));
  return 1;
}


/* ZX Spectrum exponential function to the base e

   Deviations with respect to C's exp:
   median:    1.2643013879704e-029
   mean:      1.2496836518746e-007
   maxmimum:  2.6684032491175e-005

   Reference: for arg from -100 to 10 by 0.0001

   The function is 15 % slower than C's exp function. It looses precision if its
   argument is greater than the constant `E`. */

lua_Number zxexp (lua_Number d, int gen) {
  lua_Number z, n, t1, t2, t3, t7;
  d /= LN2;  /* (D=C*(1/LN 2);EXP C=2**D). */
  n = floor(d);
  z = d - n;  /* (2**(N+Z) is now required). */
  z = 2*z - 1;
  if (gen)
    t7 = genseries(z, EXPA, EXPA_N);
  else {
    t1 = z*z;
    t2 = t1*z;
    t3 = t1*t1;
    t7 = 0.128E-6*t3*t2 + 0.59008E-4*t3*z + 0.9811784E-2*t2 + 0.49012908*z +
         0.3392E-5*t3*t1 + 0.85016E-3*t3 + 0.84932884E-1*t1 + 0.1414213563E1;
  }
  return t7 * tools_intpow(2, n);
}

static int zx_exp (lua_State *L) {
  lua_pushnumber(L, zxexp(agn_checknumber(L, 1), agnL_optboolean(L, 2, 0)));
  return 1;
}


/* Returns the ZX Spectrum exponentiation x ^ y, with x and y numbers, and returns a number.

   Internally, the ZX Spectrum and this function treats x^y like exp(ln(x)*y). If x < 0 then
   `undefined` is returned.

   Deviations with respect to C's pow:
   median:    0.00064980709794327
   mean:      1.2496836518746e-007
   maxmimum:  4437.2052326202

   Reference: for arg from 0 to 10 by 0.0001

   Large deviations with C's pow function. At least twice as slow as pow. */

lua_Number zxpow (lua_Number x, lua_Number y, int gen) {
  lua_Number r;
  if (x == 0) {
    if (y == 0) r = 1;
    else if (y > 0) r = 0;
    else r = AGN_NAN;
  } else
    r = zxexp(zxln(x, gen) * y, gen);  /* EXP (Y*LN X) */
  return r;
}

static int zx_pow (lua_State *L) {
  lua_pushnumber(L, zxpow(agn_checknumber(L, 1), agn_checknumber(L, 2), agnL_optboolean(L, 3, 0)));
  return 1;
}


/* Returns the ZX Spectrum square root of its numeric argument x and returns a number. If x < 0, `undefined`
   is returned.

   Internally, the ZX Spectrum treats sqr(x) = pow(x, 0.5).

   Deviations with respect to C's sqrt:
   median:    1.0059902155746e-007
   mean:      1.7540440159232e-007
   maxmimum:  8.5615920397686e-007

   Reference: for arg from 0 to 100 by 0.0001

   60 percent slower than C's sqrt function. */

#define zxsqr(x, gen)   zxpow(x, 0.5, gen)

static int zx_sqr (lua_State *L) {
  lua_Number x = agn_checknumber(L, 1);
  lua_pushnumber(L, zxsqr(x, agnL_optboolean(L, 3, 0)));
  return 1;
}


/* Computes the ZX Spectrum inverse tangent of its numeric argument x and returns a number.

   Deviations with respect to C's atan:
   median:    3.4858782527181e-012
   mean:      6.647332575213e-012
   maxmimum:  3.9744829649635e-010

   Reference: for arg from -100 to 100 by 0.0001

   20 percent faster than C's atan function. */

lua_Number zxatn (lua_Number b, int gen) {
  lua_Number d, z, t1, t2, t4, t6, t;
  d = (fabs(b) >= 1) ? -1/b : b;
  z = 2*d*d - 1;
  if (gen)
    t = genseries(z, ATNA, ATNA_N);
  else {
    t1 = z*z;
    t2 = t1*t1;
    t4 = t2*t2;
    t6 = t1*z;
    t = -0.3608128E-3*t2*z - 0.10187654*z - 0.22528E-5*t4*z - 0.4096E-6*t4*t6
        - 0.300032E-4*t2*t6 - 0.50309E-2*t6 + 0.1024E-5*t4*t1 + 0.84992E-5*t4
        + 0.1023936E-3*t2*t1 + 0.131556E-2*t2 + 0.208518532E-1*t1 + 0.8704197514;
  };
  t = d*t;
  if (b >= 1) t += PIO2;
  else if (b <= -1) t -= PIO2;
  return t;
}

static int zx_atn (lua_State *L) {
  lua_pushnumber(L, zxatn(agn_checknumber(L, 1), agnL_optboolean(L, 2, 0)));
  return 1;
}


/* Computes the ZX Spectrum inverse sine of its numeric argument x and returns a number.
   If x nE [-1, 1], `undefined` is returned.

   Deviations with respect to C's asin:
   median:    2.1370449854174e-009
   mean:      4.4287855517797e-009
   maxmimum:  3.0062257683205e-008

   Reference: for arg from -1 to 1 by 0.00001

   VERY slow with respect to C's asin */

lua_Number zxasn (lua_Number x, int gen) {
  lua_Number t, atn;
  t = x/(1 + zxsqr(1 - x*x, gen));  /* = tan(y/2)  */
  atn = zxatn(t, gen);
  return atn + atn;
}

static int zx_asn (lua_State *L) {
  lua_pushnumber(L, zxasn(agn_checknumber(L, 1), agnL_optboolean(L, 2, 0)));
  return 1;
}


/* Computes the ZX Spectrum inverse cosine of its numeric argument x and returns a number.
   If x nE [-1, 1], `undefined` is returned.

   Deviations with respect to C's acos:
   median:    2.1370449854174e-009
   mean:      4.4287855521004e-009
   maxmimum:  3.0062257683205e-008

   Reference: from arg from -1 to 1 by 0.00001 */

static int zx_acs (lua_State *L) {
  lua_pushnumber(L, PI/2 - zxasn(agn_checknumber(L, 1), agnL_optboolean(L, 2, 0)));
  return 1;
}


/* Just for completeness ... */


/* Returns the absolute magnitude of the number x. The function does not use Chebyshev polynomials. */

static int zx_abs (lua_State *L) {
  lua_pushnumber(L, fabs(agn_checknumber(L, 1)));
  return 1;
}


#define luai_numsign(x)     ( ((x) == 0) ? 0 : (((x) < 0) ? -1 : ((x) > 0 ? 1 : AGN_NAN)) )

/* Returns -1 if the number x is negative, 0 if x is zero, and 1 if x is positive. If x is `undefined`,
   `undefined` is returned. The function does not use Chebyshev polynomials. */

static int zx_sgn (lua_State *L) {
  lua_pushnumber(L, luai_numsign(agn_checknumber(L, 1)));
  return 1;
}


/* Rounds its numeric argument x downwards to the nearest integer. The function does not use Chebyshev polynomials. */

static int zx_int (lua_State *L) {
  lua_pushnumber(L, floor(agn_checknumber(L, 1)));
  return 1;
}


/* Returns one if its numeric argument is zero, and zero otherwise. The function does not use Chebyshev polynomials. */

static int zx_not (lua_State *L) {
  lua_pushnumber(L, agn_checknumber(L, 1) == 0);
  return 1;
}


/* Performs the binary operation 'x OR y' and returns the number x if the number y is zero,
   and the value 1 otherwise. Strings are not supported. The function does not use Chebyshev polynomials. */

static int zx_or (lua_State *L) {
  lua_Number x, y;
  x = agn_checknumber(L, 1);
  y = agn_checknumber(L, 2);
  lua_pushnumber(L,  (y == 0) ? x : 1);
  return 1;
}


/* Returns x if y is non-zero and the value zero otherwise. Strings are not supported.
   The function does not use Chebyshev polynomials. */

static int zx_and (lua_State *L) {
  lua_Number x, y;
  x = agn_checknumber(L, 1);
  y = agn_checknumber(L, 2);
  lua_pushnumber(L,  (y != 0) ? x : 0);
  return 1;
}


/* The function returns the Chebyshev coefficient vectors for various package functions. The return is a dictionary of four
   numeric sequences:

   'SIN': coefficient vector for zx.SIN and indirectly zx.COS, 6 numbers by default.
   'ATN': coefficient vector for zx.ATN and indirectly zx.ACS and zx.ASN, 12 numbers by default.
   'LN':  coefficient vector for zx.LN and indirectly zx.SQR and zx.POW, 12 numbers by default.
   'EXP': coefficient vector for zx.EXP and indirectly zx.SQR and zx.POW, 8 numbers by default. */

static int zx_getcoeffs (lua_State *L) {
  int i;
  lua_createtable(L, 0, 4);
  lua_pushstring(L, "SIN");
  agn_createseq(L, SINA_N);
  for (i=0; i < SINA_N; i++)
    lua_seqsetinumber(L, -1, i + 1, SINA[i]);
  lua_rawset(L, -3);
  lua_pushstring(L, "LN");
  agn_createseq(L, LNA_N);
  for (i=0; i < LNA_N; i++)
    lua_seqsetinumber(L, -1, i + 1, LNA[i]);
  lua_rawset(L, -3);
  lua_pushstring(L, "EXP");
  agn_createseq(L, EXPA_N);
  for (i=0; i < EXPA_N; i++)
    lua_seqsetinumber(L, -1, i + 1, EXPA[i]);
  lua_rawset(L, -3);
  lua_pushstring(L, "ATN");
  agn_createseq(L, ATNA_N);
  for (i=0; i < ATNA_N; i++)
    lua_seqsetinumber(L, -1, i + 1, ATNA[i]);
  lua_rawset(L, -3);
  return 1;
}


/* Globally sets Chebyshev coefficients to the package's environment. You can change existing coefficients, reduce or
   enlarge their respective number down to one or up to 256 values. Internally, the coefficients are treated
   as C doubles, the built-in defaults have the precision of C floats.

   The first argument must be the string 'SIN', 'ATN', 'LN', or 'EXP'. The second argument must be a sequence
   of one to 256 numbers.

   For the meanings of the first argument, see `zx.getcoeffs`.

   Please note that the respective zx functions must be called with the last argument `true` in order to revert to
   the (changed) coefficient vectors as they use hard-wired expanded polynomials by default. */

static int zx_setcoeffs (lua_State *L) {
  int i;
  size_t n;
  const char *coeff = agn_checkstring(L, 1);
  if (!lua_isseq(L, 2))
    luaL_error(L, "Error in " LUA_QS ": second argument must be a sequence, got %s.", "zx.setcoeffs", luaL_typename(L, 2));
  n = agn_seqsize(L, 2);
  if (n < 1 || n > 256)
    luaL_error(L, "Error in " LUA_QS ": one to 256 coefficients needed, got %d.", "zx.setcoeffs", n);
  if (strcmp(coeff, "SIN") == 0) {
    for (i=0; i < n; i++)
      SINA[i] = agn_seqgetinumber(L, 2, i + 1);
    SINA_N = n;
  }
  else if (strcmp(coeff, "LN") == 0) {
    for (i=0; i < n; i++)
      LNA[i] = agn_seqgetinumber(L, 2, i + 1);
    LNA_N = n;
  }
  else if (strcmp(coeff, "EXP") == 0) {
    for (i=0; i < n; i++)
      EXPA[i] = agn_seqgetinumber(L, 2, i + 1);
    EXPA_N = n;
  }
  else if (strcmp(coeff, "ATN") == 0) {
    for (i=0; i < n; i++)
      ATNA[i] = agn_seqgetinumber(L, 2, i + 1);
    ATNA_N = n;
  }
  else
    luaL_error(L, "Error in " LUA_QS ": unknown setting %s.", "zx.setcoeffs", coeff);
  return 0;
}


/* Reduces a number x to another number v in the range [-1, 1] where sin(v) = sin(Pi*v/2), and returns v. The function
   imitates the ZX Spectrum ROM 'reduce argument' Z80 assembler subroutine which is used to prepare calls to ZX Spectrum's
   sine and cosine subroutines. */
static int zx_reduce (lua_State *L) {
  lua_Number v, x, y;
  x = agn_checknumber(L, 1);
  y = x/PI2 - floor(x/PI2 + 0.5);            /* where y is greater than, or equal to, -.5 but less than +.5. */
  if (fabs(4*y) <= 1) v = 4*y;               /* V = 4*Y     if -1 <= 4*Y <= 1   - case i. */
  else if (1 < 4*y && 4*y < 2) v = 2 - 4*y;  /* V = 2-4*Y   if 1 < 4*Y < 2      - case ii. */
  else v = -4*y - 2;                         /* V = -4*Y-2  if -2 <= 4*Y < -1.  - case iii. */
  lua_pushnumber(L, v);
  return 1;
}


/* C reference functions */

static int math_sine (lua_State *L) {
  lua_pushnumber(L, sin(agn_checknumber(L, 1)));
  return 1;
}

static int math_cosine (lua_State *L) {
  lua_pushnumber(L, cos(agn_checknumber(L, 1)));
  return 1;
}

static int math_tangent (lua_State *L) {
  lua_pushnumber(L, tan(agn_checknumber(L, 1)));
  return 1;
}

static int math_loge (lua_State *L) {
  lua_pushnumber(L, log(agn_checknumber(L, 1)));
  return 1;
}

static int math_expe (lua_State *L) {
  lua_pushnumber(L, exp(agn_checknumber(L, 1)));
  return 1;
}

static int math_atn (lua_State *L) {
  lua_pushnumber(L, atan(agn_checknumber(L, 1)));
  return 1;
}

static int math_asn (lua_State *L) {
  lua_pushnumber(L, asin(agn_checknumber(L, 1)));
  return 1;
}

static int math_acs (lua_State *L) {
  lua_pushnumber(L, acos(agn_checknumber(L, 1)));
  return 1;
}

static int math_sqr (lua_State *L) {
  lua_pushnumber(L, sqrt(agn_checknumber(L, 1)));
  return 1;
}

static int math_pow (lua_State *L) {
  lua_pushnumber(L, pow(agn_checknumber(L, 1), agn_checknumber(L, 2)));
  return 1;
}


static const luaL_Reg zxlib[] = {
  {"ABS", zx_abs},                  /* added on November 15, 2015 */
  {"ACS", zx_acs},                  /* added on November 15, 2015 */
  {"AND", zx_and},                  /* added on November 15, 2015 */
  {"ASN", zx_asn},                  /* added on November 15, 2015 */
  {"ATN", zx_atn},                  /* added on November 15, 2015 */
  {"COS", zx_cos},                  /* added on November 12, 2015 */
  {"EXP", zx_exp},                  /* added on November 15, 2015 */
  {"INT", zx_int},                  /* added on November 16, 2015 */
  {"LN",  zx_ln},                   /* added on November 12, 2015 */
  {"NOT", zx_not},                  /* added on November 15, 2015 */
  {"OR",  zx_or},                   /* added on November 15, 2015 */
  {"POW", zx_pow},                  /* added on November 15, 2015 */
  {"SGN", zx_sgn},                  /* added on November 15, 2015 */
  {"SIN", zx_sin},                  /* added on November 11, 2015 */
  {"SQR", zx_sqr},                  /* added on November 15, 2015 */
  {"TAN", zx_tan},                  /* added on November 15, 2015 */
  {"genseries", zx_genseries},      /* added on November 12, 2015 */
  {"getcoeffs", zx_getcoeffs},      /* added on November 16, 2015 */
  {"reduce", zx_reduce},            /* added on November 16, 2015 */
  {"setcoeffs", zx_setcoeffs},      /* added on November 16, 2015 */
  /* C reference functions */
  {"acs", math_acs},
  {"asn", math_asn},
  {"atn", math_atn},
  {"cosine", math_cosine},
  {"expe", math_expe},
  {"loge", math_loge},
  {"pow", math_pow},
  {"sine", math_sine},
  {"sqr", math_sqr},
  {"tangent", math_tangent},
  {NULL, NULL}
};


/*
** Open ZX library
*/
LUALIB_API int luaopen_zx (lua_State *L) {
  luaL_register(L, AGENA_ZXLIBNAME, zxlib);
  lua_pushnumber(L, (float)PI);
  lua_setfield(L, -2, "PI");
  return 1;
}



/* Cosine function using Chebyshev polynomial

Maple V R4:

C0  := 0.472001216;
C1  := -0.499403258;
C2  := 0.027992080;
C3  := -0.000596695;
C4  := 0.000006704;
C5  := -0.000000047;

T0 := x -> ( 1 );
T1 := x -> ( x );
T2 := x -> ( 2 * x*x - 1 );
T3 := x -> ( 4 * x*x*x - 3 * x );
T4 := x -> ( 8 * x*x*x*x - 8 * x*x + 1 );
T5 := x -> ( 16 * x*x*x*x*x - 20 * x*x*x + 5 * x );

C0 * T0(z) + C1 * T1(z) + C2 * T2(z) + C3 * T3(z) + C4 * T4(z) + C5 * T5(z);

#define PCOS(z) \
  ( 0.444015840 - 0.497613408*z + 0.55930528e-1*z*z - 0.2385840e-2*z*z*z + 0.53632e-4*z*z*z*z - 0.752e-6*z*z*z*z*z )

static int zx_cos (lua_State *L) {
  lua_Number x, w, z, sign;
  x = agn_checknumber(L, 1)*0.15915494309189533577;  // 0.15915494309189533577 = 0.5/Pi
  x -= floor(x + 0.5);
  w = 4 * x;
  if (w > 1) { w = 2 - w; sign = -1; }
  else if (w < -1) { w = -w - 2; sign = -1; }
  else sign = 1;
  z = 2 * w * w - 1;
  lua_pushnumber(L, sign*PCOS(z));
  return 1;
} */


/*  Sine:

#define T0(x) ( 1 )
#define T1(x) ( x )
#define T2(x) ( 2 * x*x - 1 )
#define T3(x) ( 4 * x*x*x - 3 * x )
#define T4(x) ( 8 * x*x*x*x - 8 * x*x + 1 )
#define T5(x) ( 16 * x*x*x*x*x - 20 * x*x*x + 5 * x )

#define C0 1.276278962
#define C1 -.285261569
#define C2 0.009118016
#define C3 -.000136587
#define C4 0.000001185
#define C5 -.000000007

#define PSIN(z) ( C0 * T0(z) + C1 * T1(z) + C2 * T2(z) + C3 * T3(z) + C4 * T4(z) + C5 * T5(z) )
*/